﻿using System;
using System.IO;
using System.Text;
using System.Linq;
using System.Threading;
using System.Diagnostics;
using System.Windows.Forms;
using System.Collections.Generic;

using Raytracer;
using Raytracer.Math;
using Raytracer.Primitives;

using SysColor = System.Drawing.Color;

namespace Rays {
    public partial class frmMain : Form {
        public frmMain() {
            InitializeComponent();
        }

        System.Drawing.Image m_ResultImage = null;

        class BitmapRaytracer : Raytracer.Raytracer {
            public System.Drawing.Bitmap Image { get; private set; }
            public int? NotificationInterval { get; private set; }

            private int m_Count;
            private int m_LastPercentage;
            private readonly int m_Max;

            private readonly object m_Lock = new object();

            public class RaytracerProgressEventArgs : EventArgs {
                public int Percentage { get; private set; }
                public RaytracerProgressEventArgs(int percentage) { this.Percentage = percentage; }
            }

            public event EventHandler<RaytracerProgressEventArgs> Progress = null;

            public BitmapRaytracer(Scene scene,
                                   System.Drawing.Size resolution,
                                   int recursiveDepth,
                                   int? notificationInterval) : base(scene, resolution, recursiveDepth) {

                this.Image = new System.Drawing.Bitmap(resolution.Width, resolution.Height);
                this.NotificationInterval = notificationInterval;

                this.m_Max = resolution.Width * resolution.Height;
            }

            public override void SetPixel(int x, int y, SysColor col) {
                lock (m_Lock) {
                    Image.SetPixel(x, y, col);
                    if (NotificationInterval.HasValue) {
                        ++m_Count;
                        var percentage = (100 * m_Count) / m_Max;
                        if (percentage > m_LastPercentage) {
                            if (Progress != null) Progress(this, new RaytracerProgressEventArgs(percentage));
                            m_LastPercentage += NotificationInterval.Value;
                        }
                    }
                }
            }

            public override void Reset() {
                m_Count = 0;
                m_LastPercentage = -1;
            }
        }
        
        private void Render(Scene demoScene) {

            var k = 1;

            var Raytracer = new BitmapRaytracer(demoScene,
                                                new System.Drawing.Size(picPreview.Width * k, picPreview.Height * k),
                                                2,
                                                5);

            Raytracer.Progress += (sender, e) => {
                this.Invoke(new Action(() => progPreview.Value = e.Percentage), null);
            };

            var RenderDelegate = new Action(Raytracer.Render);

            var watch = new Stopwatch();
            watch.Start();

            State("Rendering ...");

            RenderDelegate.BeginInvoke(async => {
                RenderDelegate.EndInvoke(async);
                m_ResultImage = Raytracer.Image;
                this.Invoke(new Action(() => {
                    picPreview.Invalidate();
                    watch.Stop();
                    Message("Rendering completed", "Finished [{0} s]", Math.Round(watch.Elapsed.TotalSeconds, 2));
                    State("Complete");
                    previewToolStripMenuItem.Enabled = true;
                }));
            }, null);
            
        }

        private void picPreview_Paint(object sender, PaintEventArgs e) {
            if (m_ResultImage != null)
                e.Graphics.DrawImage(m_ResultImage, 0, 0, picPreview.Width, picPreview.Height);
        }

        private void Message(string cat, string msg, object arg) {
            txtEventLog.Text += string.Format(string.Format("{0}> {1}\n", cat, msg), arg);
        }

        private void State(string state) {
            lblState.Text = state;
        }

        private void previewToolStripMenuItem_Click(object sender, EventArgs e) {
            using (var Memo = new MemoryStream(Encoding.UTF8.GetBytes(txtSource.Text))) {
                var Scanner = new Scanner(Memo);
                var Parser = new Parser(Scanner);
                using (var errs = new StringWriter()) {
                    Parser.errors.errorStream = errs;
                    Parser.Parse();

                    if (Parser.errors.count == 0) {
                        Message("Loaded", "{0} errors", 0);
                        var res = Parser.ResultAst;
                        try {
                            previewToolStripMenuItem.Enabled = false;
                            var scene = SceneGenerator.EvalScene(res, new IPrimitiveFactory[] { new SphereGenerator(), new PlaneGenerator() });
                            Message("Interpretation completed", "Start rendering", "");
                            Render(scene);
                        }
                        catch (Exception ex) {
                            Message("Semantic error", "{0}", ex.Message);
                            previewToolStripMenuItem.Enabled = true;
                        }
                    }
                    else {
                        Message("Syntax error", "{0} errors" + errs.ToString(), Parser.errors.count);
                    }
                }
            }
        }

        private void frmMain_Load(object sender, EventArgs e) {
            /**/
        }

    }
}
