﻿using System;
using System.Linq;
using System.Collections.Generic;

using Rays;

using Raytracer;
using Raytracer.Math;
using Raytracer.Primitives;

using System.Reflection;

using Col = System.Drawing.Color;

namespace Rays {
    
    public interface IPrimitiveFactory {
        string Type { get; }
        Primitive Create(Ast data, Context context);
    }

    public class Context {
        private readonly Dictionary<string, Ast> m_Bindings;
        public IEnumerable<IPrimitiveFactory> Factories { get; private set; }

        public Context(Dictionary<string, Ast> bindings,
                       IEnumerable<IPrimitiveFactory> factories) {
            this.m_Bindings  = bindings;

            this.Factories = factories;
        }

        public void AddBinding(string name, Ast value) {
            m_Bindings.Add(name.ToLower(), value);
        }

        public Ast this[string index] {
            get {
                var i = index.ToLower();
                if (m_Bindings.ContainsKey(i))
                    return m_Bindings[i];
                else
                    throw new KeyNotFoundException(string.Format("Symbol #{0} cannot be found", index));
            }
        }

        public static Context CreateBase(IEnumerable<IPrimitiveFactory> factories) {
            var baseBindings = new Dictionary<string, Ast>();

            foreach (var col in new[] { Col.Red, Col.White, Col.Black, Col.Green, Col.Blue, Col.Navy, Col.Crimson, Col.Orange, Col.Yellow }) {
                var name = col.Name.ToLower();
                var value = new Vector(new[] {
                                       new Number((double)(col.R) / 255.0),
                                       new Number((double)(col.G) / 255.0),
                                       new Number((double)(col.B) / 255.0) });
                baseBindings.Add(name, value);
            }

            baseBindings.Add("checkerboard", new BuiltinMaterial(Checkerboard.Instance));

            return new Context(baseBindings, factories);
        }         
          
    }
        
    public static class SceneGenerator {
        public static T Eval<T>(this Ast expr, Context context) where T : Ast {
            if (expr is T)
                return (T)(expr);
            else 
                return context[expr.Eval<Symbol>(context).Name].Eval<T>(context);
            
            throw new InvalidCastException(string.Format("Requires type {0}, {1} given", typeof(T).Name, expr.GetType().Name));
        }

        public static Ast TryGetField(this List expr, string key, Context context) {
            foreach (var item in expr.Values) {
                var pair = item.Eval<Pair>(context);
                if (pair.Name.ToLower() == key.ToLower()) return pair.Value;
            }
            return null;
        }

        public static Ast GetField(this List expr, string key, Context context) {
            var res = expr.TryGetField(key, context);
            if (res != null)
                return res;
            else
                throw new KeyNotFoundException(string.Format("Key {0} is missing", key));
        }

        public static void Assert(this bool expr, string msg) {
            if (!expr) throw new InvalidOperationException(msg);
        }

        public static Scene EvalScene(Ast scene, IEnumerable<IPrimitiveFactory> factories) {
            Context context = Context.CreateBase(factories);

            var props = scene.Eval<List>(context);

            var background = (props.TryGetField("Background", context) ?? new Vector(new[] {
                                                                        new Number(0.0),
                                                                        new Number(0.0),
                                                                        new Number(0.0) 
                                                                     })).ToColor(context);

            var camera =  props
                         .GetField("Camera", context)
                         .ToCamera(context);

            var lights = (props.TryGetField("Lights", context) ?? new List(new Ast[] { }))
                         .Eval<List>(context)
                         .Values
                         .Select(light => light.ToLight(context))
                         .ToList();

            var materialExprs = (props.TryGetField("Materials", context) ?? new List(new Ast[] { }))
                                .Eval<List>(context)
                                .Values.Select(p => p.Eval<Pair>(context))
                                .ToList();

            foreach (var p in materialExprs) context.AddBinding(p.Name, p.Value);

            var objects = (props.TryGetField("Objects", context) ?? new List(new Ast[] { }))
                          .Eval<List>(context)
                          .Values
                          .Select(p => p.ToObject(context))
                          .ToList();

            return new Scene(background, camera, lights, objects);
        }

        public static Camera ToCamera(this Ast camera, Context context) {
            var props = camera.Eval<List>(context);
            return new Camera(props.GetField("Position", context).ToVector(context),
                              props.GetField("LookAt",   context).ToVector(context),
                              props.GetField("Width",    context).Eval<Number>(context).Value,
                              props.GetField("Height",   context).Eval<Number>(context).Value);
        }

        public static Light ToLight(this Ast light, Context context) {
            var props = light.Eval<List>(context);
            return new Light(props.GetField("Position", context).ToVector(context),
                             props.GetField("Color"   , context).ToColor(context));
        }

        public static Surface ToMaterial(this Ast material, Context context) {
            var checkedMaterial = (material is Symbol) ? context[((Symbol)material).Name] : material;

            if (checkedMaterial is BuiltinMaterial) return ((BuiltinMaterial)(checkedMaterial)).Material;
            var props = checkedMaterial.Eval<List>(context);
            return new ConstantSurface(props.GetField("Diffuse", context).ToColor(context),
                                       props.GetField("Specular", context).ToColor(context),
                                       props.GetField("Reflection", context).Eval<Number>(context).Value,
                                       props.GetField("Roughness", context).Eval<Number>(context).Value);
        }

        public static SceneObject ToObject(this Ast obj, Context context) {
            var props = obj.Eval<List>(context);
            var type = props.GetField("Type", context).Eval<Symbol>(context).Name.ToLower();

            var factory = context.Factories.FirstOrDefault(f => f.Type.ToLower() == type);
            if (factory == null) throw new InvalidOperationException(string.Format("Primitive type '{0}' cannot be created", type));
            return factory
                  .Create(obj, context)
                  .WithSurface(props
                              .GetField("Material", context)
                              .ToMaterial(context));
        }

        public static Vector3 ToVector(this Ast expr, Context context) {
            var comp = expr.Eval<Vector>(context).Components
                      .Select(x => x.Eval<Number>(context).Value)
                      .ToList();
            (comp.Count == 3).Assert("Invalid vector size");
            return new Vector3(comp[0], comp[1], comp[2]);
        }

        public static Color ToColor(this Ast expr, Context context) {
            var vec = expr.ToVector(context);
            (0.0 <= vec.x && vec.x <= 1.0).Assert("Invalid color value");
            (0.0 <= vec.y && vec.y <= 1.0).Assert("Invalid color value");
            (0.0 <= vec.z && vec.z <= 1.0).Assert("Invalid color value"); 
            return new Color(vec.x, vec.y, vec.z);
        }
    }
}